from pharmpy.results import Results


class ResmodResults(Results):
    """Resmod results class"""

    def __init__(self):
        pass


def calculate_results():
    """Calculate resmod results"""
    res = ResmodResults()
    return res


def psn_resmod_results():
    res = calculate_results()
    return res
